#!/usr/bin/env python3
"""
Tokenization Benchmark Script
Reproduces all measurements from the LLM Calculator tokenization benchmark article.

Usage:
    python benchmark.py --corpus wikipedia_en.txt --threads 16 --output results.json
    python benchmark.py --corpus python_code.txt --threads 1 --tokenizer gpt4o
    python benchmark.py --corpus chinese_wiki.txt --all-tokenizers
"""

import argparse
import json
import time
import threading
from pathlib import Path
from typing import Dict, List, Tuple, Optional
import statistics
import psutil
import os

# Core tokenization libraries
try:
    import tiktoken
    TIKTOKEN_AVAILABLE = True
except ImportError:
    TIKTOKEN_AVAILABLE = False
    print("Warning: tiktoken not available. Install with: pip install tiktoken")

try:
    from transformers import AutoTokenizer
    TRANSFORMERS_AVAILABLE = True
except ImportError:
    TRANSFORMERS_AVAILABLE = False
    print("Warning: transformers not available. Install with: pip install transformers")

try:
    import sentencepiece as spm
    SENTENCEPIECE_AVAILABLE = True
except ImportError:
    SENTENCEPIECE_AVAILABLE = False
    print("Warning: sentencepiece not available. Install with: pip install sentencepiece")


class TokenizerWrapper:
    """Wrapper for different tokenizer types to provide consistent interface."""
    
    def __init__(self, tokenizer_type: str):
        self.tokenizer_type = tokenizer_type
        self.tokenizer = None
        self.vocab_size = 0
        self.memory_usage = 0
        self._load_tokenizer()
    
    def _load_tokenizer(self):
        """Load the specified tokenizer."""
        if self.tokenizer_type == "gpt4":
            if not TIKTOKEN_AVAILABLE:
                raise ImportError("tiktoken required for GPT-4 tokenizer")
            self.tokenizer = tiktoken.get_encoding("cl100k_base")
            self.vocab_size = self.tokenizer.n_vocab
            
        elif self.tokenizer_type == "gpt4o":
            if not TIKTOKEN_AVAILABLE:
                raise ImportError("tiktoken required for GPT-4o tokenizer")
            self.tokenizer = tiktoken.get_encoding("o200k_base")
            self.vocab_size = self.tokenizer.n_vocab
            
        elif self.tokenizer_type == "llama3":
            if not TRANSFORMERS_AVAILABLE:
                raise ImportError("transformers required for Llama 3 tokenizer")
            self.tokenizer = AutoTokenizer.from_pretrained("meta-llama/Meta-Llama-3-8B")
            self.vocab_size = len(self.tokenizer.vocab)
            
        else:
            raise ValueError(f"Unknown tokenizer type: {self.tokenizer_type}")
        
        # Estimate memory usage
        self._estimate_memory()
    
    def _estimate_memory(self):
        """Estimate memory usage based on vocabulary size."""
        # Base memory: vocab_size * 4 bytes (for token IDs) + merge table
        base_memory = self.vocab_size * 4  # 4 bytes per token ID
        
        if self.tokenizer_type in ["gpt4", "gpt4o"]:
            # BPE merge table roughly 2x vocab size
            merge_table_size = self.vocab_size * 8  # Estimated
            self.memory_usage = (base_memory + merge_table_size) / (1024 * 1024)  # Convert to MB
        elif self.tokenizer_type == "llama3":
            # SentencePiece has different overhead
            self.memory_usage = (base_memory + self.vocab_size * 6) / (1024 * 1024)  # Convert to MB
    
    def encode(self, text: str) -> List[int]:
        """Encode text to tokens."""
        if self.tokenizer_type in ["gpt4", "gpt4o"]:
            return self.tokenizer.encode(text)
        elif self.tokenizer_type == "llama3":
            return self.tokenizer.encode(text)
        else:
            raise ValueError(f"Encoding not implemented for {self.tokenizer_type}")
    
    def get_info(self) -> Dict:
        """Get tokenizer information."""
        return {
            "type": self.tokenizer_type,
            "vocab_size": self.vocab_size,
            "estimated_memory_mb": round(self.memory_usage, 1)
        }


class BenchmarkRunner:
    """Main benchmark runner."""
    
    def __init__(self, corpus_path: str, threads: int = 1, warmup_runs: int = 3, test_runs: int = 10):
        self.corpus_path = Path(corpus_path)
        self.threads = threads
        self.warmup_runs = warmup_runs
        self.test_runs = test_runs
        self.results = {}
        
        # Load corpus
        self.corpus_text = self._load_corpus()
        print(f"Loaded corpus: {len(self.corpus_text):,} characters")
    
    def _load_corpus(self) -> str:
        """Load the text corpus."""
        if not self.corpus_path.exists():
            raise FileNotFoundError(f"Corpus file not found: {self.corpus_path}")
        
        with open(self.corpus_path, 'r', encoding='utf-8') as f:
            text = f.read()
        
        # Ensure minimum size for reliable benchmarking
        if len(text) < 100_000:  # 100KB minimum
            print(f"Warning: Corpus is small ({len(text):,} chars). Results may be unreliable.")
        
        return text
    
    def _single_thread_benchmark(self, tokenizer: TokenizerWrapper, text: str) -> float:
        """Benchmark tokenization on single thread."""
        start_time = time.perf_counter()
        tokens = tokenizer.encode(text)
        end_time = time.perf_counter()
        
        duration = end_time - start_time
        return len(tokens) / duration  # tokens per second
    
    def _multi_thread_benchmark(self, tokenizer: TokenizerWrapper, text: str, num_threads: int) -> float:
        """Benchmark tokenization with multiple threads."""
        # Split text into chunks
        chunk_size = len(text) // num_threads
        chunks = [text[i:i+chunk_size] for i in range(0, len(text), chunk_size)]
        
        results = []
        threads = []
        
        def worker(chunk_text: str):
            """Worker function for threading."""
            start_time = time.perf_counter()
            tokens = tokenizer.encode(chunk_text)
            end_time = time.perf_counter()
            duration = end_time - start_time
            results.append((len(tokens), duration))
        
        # Start all threads
        start_time = time.perf_counter()
        for chunk in chunks:
            t = threading.Thread(target=worker, args=(chunk,))
            t.start()
            threads.append(t)
        
        # Wait for all threads to complete
        for t in threads:
            t.join()
        
        end_time = time.perf_counter()
        total_duration = end_time - start_time
        
        # Calculate total tokens
        total_tokens = sum(r[0] for r in results)
        
        return total_tokens / total_duration
    
    def benchmark_tokenizer(self, tokenizer_type: str) -> Dict:
        """Benchmark a specific tokenizer."""
        print(f"\nBenchmarking {tokenizer_type}...")
        
        try:
            tokenizer = TokenizerWrapper(tokenizer_type)
        except Exception as e:
            print(f"Error loading {tokenizer_type}: {e}")
            return {"error": str(e)}
        
        # Warmup runs
        print(f"  Warming up ({self.warmup_runs} runs)...")
        for _ in range(self.warmup_runs):
            self._single_thread_benchmark(tokenizer, self.corpus_text[:10000])  # Small sample
        
        # Single-thread benchmark
        print(f"  Single-thread benchmark ({self.test_runs} runs)...")
        single_thread_results = []
        for run in range(self.test_runs):
            result = self._single_thread_benchmark(tokenizer, self.corpus_text)
            single_thread_results.append(result)
            print(f"    Run {run+1}: {result:,.0f} tokens/sec")
        
        # Multi-thread benchmark (if requested)
        multi_thread_results = []
        if self.threads > 1:
            print(f"  Multi-thread benchmark ({self.threads} threads, {self.test_runs} runs)...")
            for run in range(self.test_runs):
                result = self._multi_thread_benchmark(tokenizer, self.corpus_text, self.threads)
                multi_thread_results.append(result)
                print(f"    Run {run+1}: {result:,.0f} tokens/sec")
        
        # Token efficiency
        print("  Calculating token efficiency...")
        sample_text = self.corpus_text[:100000]  # 100KB sample
        tokens = tokenizer.encode(sample_text)
        tokens_per_1000_chars = (len(tokens) / len(sample_text)) * 1000
        
        # Compile results
        results = {
            "tokenizer_info": tokenizer.get_info(),
            "corpus_size_chars": len(self.corpus_text),
            "sample_size_chars": len(sample_text),
            "tokens_per_1000_chars": round(tokens_per_1000_chars, 1),
            "single_thread": {
                "mean_tokens_per_sec": round(statistics.mean(single_thread_results), 0),
                "median_tokens_per_sec": round(statistics.median(single_thread_results), 0),
                "std_dev": round(statistics.stdev(single_thread_results), 0) if len(single_thread_results) > 1 else 0,
                "all_runs": single_thread_results
            }
        }
        
        if multi_thread_results:
            scaling_factor = statistics.median(multi_thread_results) / statistics.median(single_thread_results)
            results["multi_thread"] = {
                "threads": self.threads,
                "mean_tokens_per_sec": round(statistics.mean(multi_thread_results), 0),
                "median_tokens_per_sec": round(statistics.median(multi_thread_results), 0),
                "std_dev": round(statistics.stdev(multi_thread_results), 0) if len(multi_thread_results) > 1 else 0,
                "scaling_factor": round(scaling_factor, 1),
                "all_runs": multi_thread_results
            }
        
        return results
    
    def run_benchmark(self, tokenizer_types: List[str]) -> Dict:
        """Run benchmark for all specified tokenizers."""
        print(f"Starting benchmark with {self.threads} threads on {len(self.corpus_text):,} characters")
        print(f"Hardware: {psutil.cpu_count()} CPU cores, {psutil.virtual_memory().total / (1024**3):.1f}GB RAM")
        
        all_results = {
            "benchmark_info": {
                "corpus_path": str(self.corpus_path),
                "corpus_size_chars": len(self.corpus_text),
                "threads": self.threads,
                "warmup_runs": self.warmup_runs,
                "test_runs": self.test_runs,
                "timestamp": time.strftime("%Y-%m-%d %H:%M:%S"),
                "hardware": {
                    "cpu_count": psutil.cpu_count(),
                    "memory_gb": round(psutil.virtual_memory().total / (1024**3), 1)
                }
            },
            "tokenizer_results": {}
        }
        
        for tokenizer_type in tokenizer_types:
            all_results["tokenizer_results"][tokenizer_type] = self.benchmark_tokenizer(tokenizer_type)
        
        return all_results


def create_sample_corpus(corpus_type: str, output_path: str):
    """Create sample corpus files for testing."""
    if corpus_type == "english":
        # Generate sample English text
        sample_text = """
        Artificial intelligence (AI) is intelligence demonstrated by machines, as opposed to the natural intelligence displayed by humans and animals. Leading AI textbooks define the field as the study of "intelligent agents": any device that perceives its environment and takes actions that maximize its chance of successfully achieving its goals.

        The term "artificial intelligence" was first coined in 1956 by John McCarthy at the Dartmouth Summer Research Project on Artificial Intelligence. The early days of AI were marked by optimism, with researchers believing that human-level AI was just around the corner. However, progress proved more difficult than expected, leading to several "AI winters" - periods of reduced funding and interest.

        Modern AI has achieved remarkable success in various domains including natural language processing, computer vision, game playing, and robotics. Machine learning, particularly deep learning, has been the driving force behind many recent breakthroughs. Large language models like GPT-3 and GPT-4 have demonstrated unprecedented capabilities in text generation and understanding.

        The development of AI raises important questions about ethics, safety, and the future of work. As AI systems become more capable, there is ongoing debate about how to ensure they remain aligned with human values and interests. The field of AI safety has emerged as a crucial area of research, focusing on developing AI systems that are robust, interpretable, and beneficial to humanity.
        """ * 100  # Repeat to create larger corpus
        
    elif corpus_type == "python":
        # Generate sample Python code
        sample_text = '''
def fibonacci(n):
    """Calculate the nth Fibonacci number using dynamic programming."""
    if n <= 1:
        return n
    
    dp = [0] * (n + 1)
    dp[1] = 1
    
    for i in range(2, n + 1):
        dp[i] = dp[i - 1] + dp[i - 2]
    
    return dp[n]

class BinaryTree:
    """A simple binary tree implementation."""
    
    def __init__(self, value):
        self.value = value
        self.left = None
        self.right = None
    
    def insert(self, value):
        """Insert a value into the tree."""
        if value < self.value:
            if self.left is None:
                self.left = BinaryTree(value)
            else:
                self.left.insert(value)
        else:
            if self.right is None:
                self.right = BinaryTree(value)
            else:
                self.right.insert(value)
    
    def inorder_traversal(self):
        """Perform inorder traversal of the tree."""
        result = []
        if self.left:
            result.extend(self.left.inorder_traversal())
        result.append(self.value)
        if self.right:
            result.extend(self.right.inorder_traversal())
        return result

import numpy as np
from typing import List, Dict, Optional
import json

def process_data(data: List[Dict], filter_key: str = None) -> Dict:
    """Process and analyze data with optional filtering."""
    filtered_data = data
    if filter_key:
        filtered_data = [item for item in data if filter_key in item]
    
    return {
        "count": len(filtered_data),
        "keys": list(set().union(*(d.keys() for d in filtered_data))),
        "sample": filtered_data[:5] if filtered_data else []
    }
        ''' * 50  # Repeat to create larger corpus
        
    elif corpus_type == "chinese":
        # Generate sample Chinese text
        sample_text = """
        人工智能（AI）是由机器展示的智能，与人类和动物展示的自然智能相对。领先的AI教科书将该领域定义为"智能代理"的研究：任何感知其环境并采取行动以最大化其成功实现目标机会的设备。

        "人工智能"一词最初由约翰·麦卡锡在1956年达特茅斯人工智能夏季研究项目中首次提出。AI的早期充满了乐观主义，研究人员相信人类级别的AI就在眼前。然而，进展比预期的更困难，导致了几个"AI冬天"——资金和兴趣减少的时期。

        现代AI在各个领域取得了显著成功，包括自然语言处理、计算机视觉、游戏和机器人技术。机器学习，特别是深度学习，一直是许多最新突破的推动力。像GPT-3和GPT-4这样的大型语言模型在文本生成和理解方面展现了前所未有的能力。

        AI的发展提出了关于伦理、安全和工作未来的重要问题。随着AI系统变得更加强大，关于如何确保它们与人类价值观和利益保持一致的辩论正在进行。AI安全领域已成为一个重要的研究领域，专注于开发稳健、可解释且对人类有益的AI系统。
        """ * 100  # Repeat to create larger corpus
    
    else:
        raise ValueError(f"Unknown corpus type: {corpus_type}")
    
    with open(output_path, 'w', encoding='utf-8') as f:
        f.write(sample_text)
    
    print(f"Created sample {corpus_type} corpus: {output_path} ({len(sample_text):,} characters)")


def main():
    parser = argparse.ArgumentParser(description="Tokenization benchmark tool")
    parser.add_argument("--corpus", type=str, help="Path to corpus file")
    parser.add_argument("--tokenizer", type=str, choices=["gpt4", "gpt4o", "llama3"], 
                       help="Specific tokenizer to benchmark")
    parser.add_argument("--all-tokenizers", action="store_true", 
                       help="Benchmark all available tokenizers")
    parser.add_argument("--threads", type=int, default=1, 
                       help="Number of threads for multi-threading benchmark")
    parser.add_argument("--output", type=str, default="benchmark_results.json",
                       help="Output file for results")
    parser.add_argument("--warmup-runs", type=int, default=3,
                       help="Number of warmup runs")
    parser.add_argument("--test-runs", type=int, default=10,
                       help="Number of test runs")
    parser.add_argument("--create-sample", type=str, choices=["english", "python", "chinese"],
                       help="Create sample corpus file")
    
    args = parser.parse_args()
    
    # Create sample corpus if requested
    if args.create_sample:
        output_path = f"sample_{args.create_sample}.txt"
        create_sample_corpus(args.create_sample, output_path)
        print(f"Sample corpus created: {output_path}")
        return
    
    # Validate arguments
    if not args.corpus:
        print("Error: --corpus is required (or use --create-sample to generate test data)")
        return
    
    if not args.tokenizer and not args.all_tokenizers:
        print("Error: Either --tokenizer or --all-tokenizers must be specified")
        return
    
    # Determine which tokenizers to benchmark
    if args.all_tokenizers:
        tokenizers = ["gpt4", "gpt4o", "llama3"]
    else:
        tokenizers = [args.tokenizer]
    
    # Run benchmark
    try:
        runner = BenchmarkRunner(
            corpus_path=args.corpus,
            threads=args.threads,
            warmup_runs=args.warmup_runs,
            test_runs=args.test_runs
        )
        
        results = runner.run_benchmark(tokenizers)
        
        # Save results
        with open(args.output, 'w', encoding='utf-8') as f:
            json.dump(results, f, indent=2, ensure_ascii=False)
        
        print(f"\nBenchmark complete! Results saved to: {args.output}")
        
        # Print summary
        print("\n" + "="*60)
        print("BENCHMARK SUMMARY")
        print("="*60)
        
        for tokenizer_type, result in results["tokenizer_results"].items():
            if "error" in result:
                print(f"{tokenizer_type}: ERROR - {result['error']}")
                continue
                
            print(f"\n{tokenizer_type.upper()}:")
            print(f"  Memory Usage: {result['tokenizer_info']['estimated_memory_mb']} MB")
            print(f"  Tokens per 1,000 chars: {result['tokens_per_1000_chars']}")
            print(f"  Single-thread: {result['single_thread']['median_tokens_per_sec']:,.0f} tokens/sec")
            
            if "multi_thread" in result:
                print(f"  Multi-thread ({result['multi_thread']['threads']} threads): {result['multi_thread']['median_tokens_per_sec']:,.0f} tokens/sec")
                print(f"  Scaling factor: {result['multi_thread']['scaling_factor']}x")
        
    except Exception as e:
        print(f"Error running benchmark: {e}")
        import traceback
        traceback.print_exc()


if __name__ == "__main__":
    main() 